// SPDX-FileCopyrightText: 2022 Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: AGPL-3.0-or-later

use {
    crate::api_client::ApiClient,
    crate::commands::import,
    crate::constants,
    crate::event::{Event, Meal},
    crate::recipe,
    anyhow::{bail, Result},
    chrono::naive::NaiveDate,
    futures::future::try_join_all,
    reqwest::StatusCode,
    std::collections::{HashMap, HashSet},
    std::fmt::Write,
    std::iter::Iterator,
    std::path::Path,
    std::rc::Rc,
};

#[derive(serde::Deserialize)]
struct CsvRecord {
    day: NaiveDate,
    lunch: String,
    dinner: String,
}

pub async fn with(api_client: &ApiClient, calendar: &str, csv_file: &Path) -> Result<()> {
    let mut csv = csv::Reader::from_path(csv_file)?;
    let records = csv.deserialize::<CsvRecord>().flatten().collect::<Vec<_>>();

    let recipe_urls = urls_from_csv(records.iter())?;
    import::with(&api_client, recipe_urls.into_iter()).await?;
    let recipes = get_all_recipes(&api_client).await?;

    let events = records
        .iter()
        .flat_map(|r| {
            let lunch = recipes.get(&r.lunch);
            let dinner = recipes.get(&r.dinner);

            let events = [
                lunch.map(|recipe| Event::new(r.day, Meal::Lunch, recipe.clone())),
                dinner.map(|recipe| Event::new(r.day, Meal::Dinner, recipe.clone())),
            ];

            events
        })
        .flatten();

    publish_events(&api_client, calendar, events).await?;
    Ok(())
}

fn urls_from_csv<'a, RecordsIter>(records: RecordsIter) -> Result<HashSet<String>>
where
    RecordsIter: Iterator<Item = &'a CsvRecord>,
{
    Ok(
        records.fold(std::collections::HashSet::new(), |mut set, r| {
            if !r.lunch.is_empty() {
                set.insert(r.lunch.clone());
            }
            if !r.dinner.is_empty() {
                set.insert(r.dinner.clone());
            }
            set
        }),
    )
}

async fn get_all_recipes(api_client: &ApiClient) -> Result<HashMap<String, Rc<recipe::Recipe>>> {
    let metadata = api_client
        .rest()
        .get(api_client.base_url().join("apps/cookbook/api/recipes")?)
        .send()
        .await?
        .json::<Vec<recipe::Metadata>>()
        .await?;

    let recipes = metadata.iter().map(|rm| async {
        let response = api_client
            .rest()
            .get(
                api_client
                    .base_url()
                    .join(&format!("apps/cookbook/api/recipes/{id}", id = rm.id))
                    .unwrap(),
            )
            .send()
            .await
            .expect(&format!(
                "Cannot fetch recipe {} with id {}",
                rm.name, rm.id
            ));
        response.json::<recipe::Recipe>().await.map(|r| Rc::new(r))
    });

    let recipes = try_join_all(recipes).await?;
    Ok(HashMap::from_iter(
        recipes.into_iter().map(|r| (r.url.clone(), r)),
    ))
}

async fn publish_events<'a, EventsIter>(
    api_client: &ApiClient,
    calendar: &str,
    events: EventsIter,
) -> Result<()>
where
    EventsIter: Iterator<Item = Event>,
{
    let calendar_prototype: ics::ICalendar = ics::ICalendar::new(
        "2.0",
        format!(
            "-//IDN {}//{} {}//EN",
            constants::VENDOR,
            env!("CARGO_PKG_NAME"),
            env!("CARGO_PKG_VERSION")
        ),
    );

    let dav_base = api_client
        .rest()
        .head(api_client.base_url().join("/.well-known/caldav")?)
        .send()
        .await?;

    let calendar_url = dav_base.url().join(&format!(
        "calendars/{}/{}/",
        &api_client.username(),
        calendar.to_lowercase().as_str().replace(" ", "-")
    ))?;

    let calendar_prototype = &calendar_prototype;
    let calendar_url = &calendar_url;
    let update_requests = events.map(|ev| async move {
        let url = calendar_url.join(&format!("{}.ics", ev.uid)).unwrap();
        let mut cal = calendar_prototype.clone();
        cal.add_event(ev.into());

        api_client
            .rest()
            .put(url)
            .header("Content-Type", "text/calendar; charset=utf-8")
            .body(cal.to_string())
            .send()
            .await
    });

    let responses = try_join_all(update_requests).await?;
    let failed_responses = responses.into_iter().filter(|response| {
        ![StatusCode::NO_CONTENT, StatusCode::CREATED].contains(&response.status())
    });

    let mut errors = String::new();
    for r in failed_responses {
        write!(errors, "\n{}", r.text().await.unwrap())?;
    }

    if !errors.is_empty() {
        bail!("Error while updating calendar events: {}", errors);
    }

    Ok(())
}
